/**
 *  Demonstration of mutexes, semaphores, and conditional variables
 *  in SDL, by Tyler Montbriand, 2003.
 *
 *  This code is freeware.  Do whatever you want with it.
 *
 *  This is a simple demonstration of semaphores, mutexes, and 
 *  conditionals.  There didn't seem to be a good tutorial for conds,
 *  so I somehow hashed out this example and in the process figured
 *  out just what the hell they are.
 *
 *  Mutexes are simple.  Mutex is short for 'Mutal Exclusion', as in
 *  only one thing at a time can lock it.  If 2 threads are calling
 *  SDL_mutexP() on one mutex, one thread will continue while the other 
 *  thread will suspend until the first calls SDL_mutexP(), at which
 *  the first will be able to lock it.  Works for any number of threads.
 *
 *  Semaphores are a little more complicated, but not much.  They are
 *  thread-safe integers with only two operations, increment
 *  ( SDL_SemPost() ) and decrement ( SDL_SemWait() ).  Increment calls 
 *  never block any thread from running, but decrement calls can:  If 
 *  you call SDL_SemWait() when the semaphore's value is 0, that thread
 *  will halt until something else increments it.  The semaphore's value
 *  thus never goes below zero.  In fact, semaphores and mutexes are  
 *  much the same.  You could build a mutex out of a semaphore.
 *
 *  Now for conditionals:  In conjunction with one mutex, these let you 
 *  control many threads.  This is the sequence that seems to work:
 *
 *  1) Thread locks mutex, and calls SDL_CondWait.  This suspends the 
 *     thread and unlocks the mutex.  
 *  2) Repeat #1 for as many threads as you like.
 *  3) Controlling program, when all threads are ready, locks mutex,
 *     calls SDL_CondBroadcast, then unlocks mutex.
 *  4) First thread is restarted.
 *  5) Second thread is restarted when first unlocks mutex.
 *  6) Repeat for as many threads as you like.
 *
 *  SDL_CondSignal differs in that it restarts one thread instead of all 
 *  of them.
 */
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <stdlib.h>

SDL_sem   *sem=NULL;   /* Global semaphore to track thread states */
SDL_mutex *mutex=NULL; /* Global mutex and conditional to launch threads */
SDL_cond  *launch=NULL;

/* Function that will be executed by the threads */
int zig(void *t);

#define ZIGS 4

int main(int argc, char *argv[])
{
  long int n;
  SDL_Thread *t[ZIGS];

  if(SDL_Init(SDL_INIT_EVENTTHREAD)<0)
  {
    fprintf(stderr,"Couldn't init SDL\n");
    return(1);
  }

  mutex=SDL_CreateMutex();
  sem=SDL_CreateSemaphore(0);
  launch=SDL_CreateCond();

  if(mutex && launch && sem)
  {
    fprintf(stderr,"main] Take off every zig!\n");

    // Create all threads
    for(n=0; n<ZIGS; n++)
      t[n]=SDL_CreateThread(zig,(void *)(long)n);

    /* wait until every zig has locked the mutex */
    for(n=0; n<ZIGS; n++)
      if(t[n]!=NULL) SDL_SemWait(sem);

    /* At this point we know that every thread has locked the mutex
       then called SDL_SemPost, and from that we know that every thread
       but the last one has unlocked the mutex with an SDL_CondWait 
       call.

       The last thread might still be running, or might not.
       main() thus attempts to lock the mutex, suspending itself until 
       the last thread is ready, then unlocks it immediately after.
     */
    SDL_mutexP(mutex);
    SDL_mutexV(mutex);

    fprintf(stderr,"main] MOVE ZIG!\n");
    SDL_CondBroadcast(launch); /* srow ze svitch! */

    for(n=0; n<ZIGS; n++) /* Wait until all threads have finished */
      if(t[n]!=NULL) SDL_WaitThread(t[n],NULL);

    fprintf(stderr,"main] You know what you doing\n");
  }
  else
    fprintf(stderr,"Couldn't create signalling devices\n");

  if(sem!=NULL)    SDL_DestroySemaphore(sem);
  if(mutex!=NULL)  SDL_DestroyMutex(mutex);
  if(launch!=NULL) SDL_DestroyCond(launch);
  return(0);
}

int zig(void *t)
{
  if((mutex==NULL)||(launch==NULL)||(sem==NULL)) return(-1);

  SDL_mutexP(mutex); /* Need to lock mutex before calling SDL_CondWait */
  {
    fprintf(stderr,"%02d] Zig ready for launch.\n",(Sint32)(long)t);

    SDL_SemPost(sem); /* Notify main that we have locked mutex */

    SDL_CondWait(launch,mutex); /* Wait until main triggers all threads */

    /* Critical section - no other threads are running in this part */
    { 
      fprintf(stderr,"%02d] Zig launched!\n",(Sint32)(long)t);
      SDL_Delay(500);
    }
  }
  SDL_mutexV(mutex); /* exit critical section - let next zig launch */

  SDL_Delay(1000); /* Can happen the same time as other threads */

  fprintf(stderr,"%02d] Zig done.\n",(Sint32)(long)t);

  return(0);
}
